import itertools

import numpy as np
import pandas as pd
import pytest
import tensorflow as tf

columns = ["ref", "x1", "x2"]

seed = 1
tf.random.set_seed(seed)
np.random.seed(seed)


def gen_value(column, line, id=0):
    if column == "ref":
        return id * 1e5 + line
    else:
        return np.random.randint(0, 1000)


def gen_df(columns, date_range, id=0):
    periods = date_range.size
    df = pd.DataFrame(
        {
            "date_time": date_range,
            **{
                col: [gen_value(col, line, id) for line in range(periods)]
                for col in columns
            },
        }
    )
    return df


def gen_df_with_id(ids, columns, date_range):
    dfs = []
    for i in ids:
        df = gen_df(columns, date_range, i)
        df["id"] = i
        dfs.append(df)
    df = pd.concat(dfs)

    return df


@pytest.fixture(
    scope="function",
    params=[(columns, 48 * 30 * 3)],
)
def time_series_df(request):
    df = gen_df(
        columns=request.param[0],
        date_range=pd.date_range("1/1/1", periods=request.param[1], freq="30min"),
    )
    return df


@pytest.fixture(
    scope="function",
    params=[
        (list(range(5)), columns, 48 * 30 * 3),
    ],
)
def time_series_df_with_id(request):
    ids, columns, periods = request.param
    df = gen_df_with_id(
        ids=ids,
        columns=columns,
        date_range=pd.date_range("1/1/1", periods=periods, freq="30min"),
    )
    return df


@pytest.fixture(scope="function")
def tmp_csv(tmpdir_factory, time_series_df):
    file_path = tmpdir_factory.mktemp("csv_data") / "test.csv"

    time_series_df.to_csv(file_path, index=False)

    return file_path, time_series_df


@pytest.fixture(scope="function")
def tmp_csv_with_id(tmpdir_factory, time_series_df_with_id):
    file_path = tmpdir_factory.mktemp("csv_data") / "test.csv"

    time_series_df_with_id.to_csv(file_path, index=False)

    return file_path, time_series_df_with_id


@pytest.fixture(params=[0, 1, 48])
def history_size(request):
    return request.param


@pytest.fixture
def prediction_size(history_size):
    return history_size


@pytest.fixture
def shift(prediction_size):
    return prediction_size


@pytest.fixture(params=[32])
def batch_size(request):
    return request.param


@pytest.fixture(params=[[], ["ref"], ["x2", "ref"], columns])
def history_columns(request):
    return request.param


@pytest.fixture(params=[[], ["ref", "x1"]])
def meta_columns(request):
    return request.param


@pytest.fixture(
    params=[
        [],
        list(
            itertools.chain(
                ["ref"],
                *[
                    [c + "_sin", c + "_cos"]
                    for c in ["weekday", "dayofyear", "time", "month"]
                ],
            )
        ),
    ]
)
def meta_columns_cycle(request):
    return request.param


@pytest.fixture
def prediction_columns(history_columns):
    return history_columns



import numpy as np
import pandas as pd
import pytest

from tensorflow_time_series_dataset.preprocessors import (
    CyclicalFeatureEncoder,
    GroupbyDatasetGenerator,
    TimeSeriesSplit,
)


@pytest.fixture
def cycl_df():
    date_range = pd.date_range(start="1/1/1992", end="31/12/1992", freq="30min")
    test_df = pd.DataFrame(
        index=date_range,
        data=np.stack(
            [
                np.random.randint(0, 10, date_range.size),
                np.random.randint(20, 100, date_range.size),
            ],
            axis=1,
        ),
        columns=["x1", "x2"],
    )
    return test_df


def test_cyclical_data_encoder(cycl_df):
    encs = {
        "x1": dict(cycl_max=9),
        "x2": dict(cycl_max=99, cycl_min=20),
        "weekday": dict(cycl_max=6),
        "dayofyear": dict(cycl_max=366, cycl_min=1),
        "month": dict(cycl_max=12, cycl_min=1),
        "time": dict(
            cycl_max=24 * 60 - 1,
            cycl_getter=lambda df, k: df.index.hour * 60 + df.index.minute,
        ),
    }
    for name, kwargs in encs.items():
        enc = CyclicalFeatureEncoder(name, **kwargs)
        enc_dat = enc(cycl_df)
        cycl = enc.cycl_getter(cycl_df, name)
        assert np.isclose(
            enc.decode(enc_dat[name + "_sin"], enc_dat[name + "_cos"]), cycl
        ).all(), "Decoding failed"


def test_cyclical_data_encoder_except(cycl_df):
    encs = {
        "x1": dict(cycl_max=cycl_df.x1.max() - 1),
        "weekday": dict(cycl_max=cycl_df.index.weekday.max() - 1),
        "dayofyear": dict(cycl_max=cycl_df.index.dayofyear.max() - 1),
        "month": dict(cycl_max=cycl_df.index.month.max() - 1),
        "time": dict(
            cycl_max=(cycl_df.index.hour * 60 + cycl_df.index.minute).max() - 1,
            cycl_getter=lambda df, k: df.index.hour * 60 + df.index.minute,
        ),
    }
    for name, kwargs in encs.items():
        enc = CyclicalFeatureEncoder(name, **kwargs)
        with pytest.raises(AssertionError):
            enc(cycl_df)


def test_time_series_split(time_series_df_with_id):
    df = time_series_df_with_id.set_index("date_time")
    l_splitter = TimeSeriesSplit(0.5, TimeSeriesSplit.LEFT)
    r_splitter = TimeSeriesSplit(0.5, TimeSeriesSplit.RIGHT)
    l_split = l_splitter(df)
    r_split = r_splitter(df)
    assert not l_split.index.isin(r_split.index).any(), "Splits overlap"
    assert l_split.index.max() < r_split.index.min(), "wrong split"
    assert np.unique(l_split.index.date).size == np.floor(
        np.unique(df.index.date).size / 2
    ) and np.unique(r_split.index.date).size == np.ceil(
        np.unique(df.index.date).size / 2
    ), "Unexpected split size"
    assert (r_split.groupby([r_split.index.date, "id"]).x1.count() == 48).all() and (
        l_split.groupby([l_split.index.date, "id"]).x1.count() == 48
    ).all(), "Incomplete Days in Split"


def test_groupby_dataset_generator(time_series_df_with_id):
    df = time_series_df_with_id.set_index("date_time")
    columns = list(sorted(df.columns))
    df = df[columns]

    records_per_id = df.groupby("id").size().max()
    expected_shape = (records_per_id, len(columns))

    idx_from_column = {c: i for i, c in enumerate(columns)}
    gen = GroupbyDatasetGenerator("id", columns=columns)
    for d in gen(df.sample(frac=1)):
        assert d.shape == expected_shape, "Wrong shape"
        id = int(d[0, idx_from_column["ref"]] // 1e5)
        expected_values = df[df.id == id].sort_index().values
        assert np.all(
            d.numpy() == expected_values,
        ), "Error: DatasetGenerator failed"



from tensorflow_time_series_dataset.factory import WindowedTimeSeriesDatasetFactory
from tensorflow_time_series_dataset.loaders import CSVDataLoader
from tensorflow_time_series_dataset.preprocessors import (
    CyclicalFeatureEncoder,
    GroupbyDatasetGenerator,
    TimeSeriesSplit,
)
from tensorflow_time_series_dataset.utils.test import validate_args, validate_dataset


def test_windowed_time_series_dataset_factory(
    time_series_df,
    batch_size,
    history_size,
    prediction_size,
    shift,
    history_columns,
    meta_columns,
    prediction_columns,
):
    df = time_series_df.set_index("date_time")

    common_kwargs = dict(
        history_size=history_size,
        prediction_size=prediction_size,
        shift=shift,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
        batch_size=batch_size,
        drop_remainder=False,
    )
    factory_kwargs = dict(
        cycle_length=1,
        shuffle_buffer_size=100,
    )

    with validate_args(
        history_size=history_size,
        prediction_size=prediction_size,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
    ):
        factory = WindowedTimeSeriesDatasetFactory(**common_kwargs, **factory_kwargs)
        ds = factory(df)
        validate_dataset(
            df,
            ds,
            **common_kwargs,
        )


def test_windowed_time_series_dataset_factory_groupby(
    time_series_df_with_id,
    batch_size,
    history_size,
    prediction_size,
    shift,
    history_columns,
    meta_columns,
    prediction_columns,
):
    df = time_series_df_with_id.set_index("date_time")

    ids = df.id.unique()
    common_kwargs = dict(
        history_size=history_size,
        prediction_size=prediction_size,
        shift=shift,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
        batch_size=batch_size,
        drop_remainder=True,
    )
    factory_kwargs = dict(
        cycle_length=len(ids),
        shuffle_buffer_size=100,
    )

    with validate_args(
        history_size=history_size,
        prediction_size=prediction_size,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
    ):
        factory = WindowedTimeSeriesDatasetFactory(**common_kwargs, **factory_kwargs)
        factory.add_preprocessor(
            GroupbyDatasetGenerator(
                "id", columns=history_columns + meta_columns + prediction_columns
            )
        )

        ds = factory(df)
        validate_dataset(
            df,
            ds,
            **common_kwargs,
        )


def test_windowed_time_series_dataset_factory_csv_loader(
    tmp_csv,
    batch_size,
    history_size,
    prediction_size,
    shift,
    history_columns,
    meta_columns,
    prediction_columns,
):
    test_data_path, df = tmp_csv
    df = df.set_index("date_time")

    common_kwargs = dict(
        history_size=history_size,
        prediction_size=prediction_size,
        shift=shift,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
        batch_size=batch_size,
        drop_remainder=False,
    )
    factory_kwargs = dict(
        cycle_length=1,
        shuffle_buffer_size=100,
    )
    with validate_args(
        history_size=history_size,
        prediction_size=prediction_size,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
    ):
        factory = WindowedTimeSeriesDatasetFactory(**common_kwargs, **factory_kwargs)
        factory.set_data_loader(CSVDataLoader(file_path=test_data_path))
        ds = factory()
        validate_dataset(
            df,
            ds,
            **common_kwargs,
        )


def test_windowed_time_series_dataset_factory_csv_loader_with_preprocessors(
    tmp_csv_with_id,
    batch_size,
    history_size,
    prediction_size,
    shift,
    history_columns,
    meta_columns_cycle,
    prediction_columns,
):
    # define encoder args
    # [name: kwargs]
    encs = {
        "weekday": dict(cycl_max=6),
        "dayofyear": dict(cycl_max=366, cycl_min=1),
        "month": dict(cycl_max=12, cycl_min=1),
        "time": dict(
            cycl_max=24 * 60 - 1,
            cycl_getter=lambda df, k: df.index.hour * 60 + df.index.minute,
        ),
    }
    test_data_path, df = tmp_csv_with_id
    test_df = df.set_index("date_time")
    for name, kwargs in encs.items():
        enc = CyclicalFeatureEncoder(name, **kwargs)
        test_df = enc(test_df)
    splitter = TimeSeriesSplit(0.5, TimeSeriesSplit.LEFT)
    test_df = splitter(test_df)

    ids = df.id.unique()

    common_kwargs = dict(
        history_size=history_size,
        prediction_size=prediction_size,
        shift=shift,
        history_columns=history_columns,
        meta_columns=meta_columns_cycle,
        prediction_columns=prediction_columns,
        batch_size=batch_size,
        drop_remainder=True,
    )
    factory_kwargs = dict(
        cycle_length=len(ids),
        shuffle_buffer_size=100,
    )
    with validate_args(
        history_size=history_size,
        prediction_size=prediction_size,
        history_columns=history_columns,
        meta_columns=meta_columns_cycle,
        prediction_columns=prediction_columns,
    ):
        factory = WindowedTimeSeriesDatasetFactory(**common_kwargs, **factory_kwargs)
        factory.set_data_loader(CSVDataLoader(file_path=test_data_path))
        factory.add_preprocessor(splitter)
        for name, kwargs in encs.items():
            factory.add_preprocessor(CyclicalFeatureEncoder(name, **kwargs))
        factory.add_preprocessor(
            GroupbyDatasetGenerator(
                "id", columns=history_columns + meta_columns_cycle + prediction_columns
            )
        )
        ds = factory()
        validate_dataset(
            test_df,
            ds,
            **common_kwargs,
        )



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : csv_data_loader.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 13:01:07 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
from typing import Union

import pandas as pd


def _read_csv_file(
    file_path: str, date_time_col: str = "date_time", **kwargs: Union[str, bool]
) -> pd.DataFrame:
    """Read CSV file into a pandas DataFrame.

    Parameters
    ----------
    file_path : str
        File path to the CSV file.
    date_time_col : str, optional
        Name of the datetime column in the CSV file.
    **kwargs
        Additional keyword arguments for pd.read_csv.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the data from the CSV file.

    Raises
    ------
    ValueError
        If the data contains NaN values.

    """
    load_data = pd.read_csv(
        file_path,
        parse_dates=[date_time_col],
        index_col=[date_time_col],
        **kwargs,
    )

    if load_data.isnull().any().sum() != 0:
        raise ValueError("Data contains NaNs")

    return load_data


class CSVDataLoader:
    """Load data from a CSV file.

    Parameters
    ----------
    file_path : str
        File path to the CSV file.
    **kwargs
        Additional keyword arguments for pd.read_csv.

    """

    def __init__(self, file_path: str, **kwargs: Union[str, bool]):
        self.file_path = file_path
        self.kwargs = kwargs

    def __call__(self) -> pd.DataFrame:
        """Load data from the CSV file using _read_csv_file.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the data from the CSV file.

        """
        return _read_csv_file(self.file_path, **self.kwargs)



import numpy as np
import pytest
import tensorflow as tf

from tensorflow_time_series_dataset.pipeline.patch_generator import PatchGenerator
from tensorflow_time_series_dataset.pipeline.patch_processor import PatchPreprocessor
from tensorflow_time_series_dataset.pipeline.windowed_time_series_pipeline import (
    WindowedTimeSeriesPipeline,
)
from tensorflow_time_series_dataset.preprocessors.groupby_dataset_generator import (
    GroupbyDatasetGenerator,
)
from tensorflow_time_series_dataset.utils.test import validate_args, validate_dataset

# FIXTURES ####################################################################


@pytest.fixture
def groupby_dataset(
    time_series_df_with_id, history_columns, meta_columns, prediction_columns
):
    time_series_df_with_id.set_index("date_time", inplace=True)
    used_cols = set(history_columns + meta_columns + prediction_columns)
    gen = GroupbyDatasetGenerator("id", columns=used_cols)
    return gen(time_series_df_with_id), time_series_df_with_id[list(used_cols) + ["id"]]


@pytest.fixture
def patched_dataset(
    request, time_series_df, history_columns, meta_columns, prediction_columns
):
    window_size, shift = request.param
    df = time_series_df.set_index("date_time")
    used_cols = set(history_columns + meta_columns + prediction_columns)

    ds = tf.data.Dataset.from_tensors(df[sorted(used_cols)])
    ds = ds.interleave(
        PatchGenerator(window_size=window_size, shift=shift, filter_nans=False),
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )
    return ds, df, window_size, shift


# TESTS #######################################################################
@pytest.mark.parametrize("window_size,shift", [(2 * 48, 48), (48 + 1, 1)])
def test_patch_generator(time_series_df, window_size, shift):
    df = time_series_df.set_index("date_time")

    initial_size = window_size - shift
    data_size = df.index.size - initial_size
    patches = data_size // shift

    expected_shape = (window_size, len(df.columns))

    ds = tf.data.Dataset.from_tensors(df)
    ds_patched = ds.interleave(
        PatchGenerator(window_size=window_size, shift=shift, filter_nans=False),
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )
    for i, patch in enumerate(ds_patched.as_numpy_iterator()):
        assert patch.shape == expected_shape, "Wrong shape"
        x1 = patch[0, 0]
        idx = int(x1 % 1e5)
        expected_values = df.iloc[idx : idx + window_size]
        assert np.all(patch == expected_values), "Patch contains wrong data"
    assert i + 1 == patches, "Not enough patches"


@pytest.mark.parametrize("window_size,shift", [(2 * 48, 48), (48 + 1, 1)])
def test_patch_generator_filter_nans(time_series_df, window_size, shift):
    df = time_series_df.set_index("date_time")
    # randomly set 20% of elemnts in the dataset for nans

    df = time_series_df.set_index("date_time")
    nan_mask = np.random.default_rng(1).uniform(0, 1, df.shape) < 0.01
    df[nan_mask] = np.nan

    initial_size = window_size - shift
    data_size = df.index.size - initial_size
    patches = data_size // shift

    expected_shape = (window_size, len(df.columns))

    ds = tf.data.Dataset.from_tensors(df)
    ds_patched = ds.interleave(
        PatchGenerator(window_size=window_size, shift=shift, filter_nans=True),
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )
    for i, patch in enumerate(ds_patched.as_numpy_iterator()):
        assert patch.shape == expected_shape, "Wrong shape"
        x1 = patch[0, 0]
        idx = int(x1 % 1e5)
        expected_values = df.iloc[idx : idx + window_size]
        assert np.all(patch == expected_values), "Patch contains wrong data"
        assert not np.isnan(patch).any(), "Patch contains NaNs."
    assert i + 1 < patches, "No patches have been dropped"


@pytest.mark.parametrize("window_size,shift", [(2 * 48, 48), (48 + 1, 1)])
def test_patch_generator_groupby(groupby_dataset, window_size, shift):
    ds, df = groupby_dataset
    records_per_id = df.groupby("id").size().max()
    ids = df.id.unique()
    columns = sorted([c for c in df.columns if c != "id"])

    initial_size = window_size - shift
    data_size = records_per_id - initial_size
    patches = data_size / shift * len(ids)
    expected_shape = (window_size, len(columns))

    ds_patched = ds.interleave(
        PatchGenerator(window_size=window_size, shift=shift, filter_nans=True),
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )

    for i, patch in enumerate(ds_patched.as_numpy_iterator()):
        assert patch.shape == expected_shape, "Wrong shape"
        if len(columns):
            x1 = patch[0, 0]
            id = int(x1 // 1e5)
            idx = int(x1 % 1e5)
            expected_values = df[df.id == id].iloc[idx : idx + window_size]
            assert np.all(
                patch == expected_values[columns].values
            ), "Patch contains wrong data"
    assert i + 1 == patches, "Not enough patches"


@pytest.mark.parametrize(
    "patched_dataset", [(2 * 48, 48), (48 + 1, 1)], indirect=["patched_dataset"]
)
def test_batch_processor(
    patched_dataset,
    history_size,
    batch_size,
    history_columns,
    meta_columns,
    prediction_columns,
):
    ds, df, window_size, shift = patched_dataset
    prediction_size = window_size - history_size

    batch_kwargs = dict(
        history_size=history_size,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
    )

    with validate_args(
        history_size=history_size,
        prediction_size=prediction_size,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
    ):
        ds_batched = ds.map(
            PatchPreprocessor(**batch_kwargs),
            num_parallel_calls=tf.data.experimental.AUTOTUNE,
        ).batch(batch_size)

        validate_dataset(
            df,
            ds_batched,
            batch_size=batch_size,
            prediction_size=prediction_size,
            shift=shift,
            drop_remainder=False,
            **batch_kwargs,
        )


def test_windowed_time_series_pipeline(
    time_series_df,
    batch_size,
    history_size,
    prediction_size,
    shift,
    history_columns,
    meta_columns,
    prediction_columns,
):
    df = time_series_df.set_index("date_time")
    used_cols = sorted(
        set(
            history_columns + meta_columns + prediction_columns,
        )
    )

    batch_kwargs = dict(
        history_size=history_size,
        prediction_size=prediction_size,
        shift=shift,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
        batch_size=batch_size,
        drop_remainder=True,
    )
    pipeline_kwargs = dict(
        cycle_length=1, shuffle_buffer_size=100, cache=True, filter_nans=False
    )

    with validate_args(
        history_size=history_size,
        prediction_size=prediction_size,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
    ):
        pipeline = WindowedTimeSeriesPipeline(**batch_kwargs, **pipeline_kwargs)
        ds = tf.data.Dataset.from_tensors(df[used_cols])
        ds = pipeline(ds)
        validate_dataset(
            df,
            ds,
            **batch_kwargs,
        )


def test_windowed_time_series_pipeline_groupby(
    groupby_dataset,
    batch_size,
    history_size,
    prediction_size,
    shift,
    history_columns,
    meta_columns,
    prediction_columns,
):
    ds, df = groupby_dataset

    ids = df.id.unique()

    batch_kwargs = dict(
        history_size=history_size,
        prediction_size=prediction_size,
        shift=shift,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
        batch_size=batch_size,
        drop_remainder=False,
    )
    pipeline_kwargs = dict(
        cycle_length=len(ids), shuffle_buffer_size=1000, cache=True, filter_nans=False
    )

    with validate_args(
        history_size=history_size,
        prediction_size=prediction_size,
        history_columns=history_columns,
        meta_columns=meta_columns,
        prediction_columns=prediction_columns,
    ):
        pipeline = WindowedTimeSeriesPipeline(**batch_kwargs, **pipeline_kwargs)
        ds = pipeline(ds)
        validate_dataset(
            df,
            ds,
            **batch_kwargs,
        )



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : test.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 13:09:50 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
from contextlib import nullcontext
from typing import ContextManager, List, Union

import numpy as np
import pytest
from numpy.typing import ArrayLike
from pandas import DataFrame


def get_idx(ref: ArrayLike, val: float) -> int:
    """Get the index of a value in a reference array.

    Parameters
    ----------
    ref : ArrayLike
        Reference array to search.
    val : float
        Value to find in the reference array.

    Returns
    -------
    int
        Index of the value in the reference array.

    """
    idx = np.where(ref == val)[0]
    assert len(idx) == 1, "Could not determine index from reference value"
    return int(idx)


def gen_patch(df: DataFrame, idx: int, size: int) -> np.ndarray:
    """Generate a patch of data from a DataFrame.

    Parameters
    ----------
    df : DataFrame
        DataFrame containing the data.
    idx : int
        Starting index of the patch.
    size : int
        Size of the patch to generate.

    Returns
    -------
    np.ndarray
        Patch of data from the DataFrame.

    """
    x = df.values[idx : idx + size]
    assert len(x) == size, "Could not generate patch from reference data"
    return x


def gen_batch(
    df: DataFrame, columns: list, size: int, refs: ArrayLike, ref_col: str
) -> np.ndarray:
    """Generate a batch of data from a DataFrame based on reference values.

    Parameters
    ----------
    df : DataFrame
        DataFrame containing the data.
    columns : list
        List of columns to include in the batch.
    size : int
        Size of each patch in the batch.
    refs : ArrayLike
        Array of reference values.
    ref_col : str
        Name of the column containing reference values.

    Returns
    -------
    np.ndarray
        Batch of data based on reference values.

    """
    batch = []
    columns = list(sorted(columns))
    for ref in refs:
        if "id" in df.columns:
            id = ref // 1e5
            idx = get_idx(df[df.id == id][ref_col], ref)
            p = gen_patch(df[df.id == id][columns], idx, size)
        else:
            idx = get_idx(df[ref_col], ref)
            p = gen_patch(df[columns], idx, size)
        batch.append(p)
    return np.float32(batch)


def validate_dataset(
    df: DataFrame,
    ds: object,
    batch_size: int,
    history_size: int,
    prediction_size: int,
    shift: int,
    history_columns: list,
    meta_columns: list,
    prediction_columns: list,
    drop_remainder: bool,
    history_reference_column: str = "ref",
    meta_reference_column: str = "ref",
    prediction_reference_column: str = "ref",
    **kwargs,
) -> int:
    """Validate the dataset generator with the given parameters.

    Parameters
    ----------
    df : DataFrame
        DataFrame containing the data.
    ds : object
        Dataset object used for validation.
    batch_size : int
        Size of each batch.
    history_size : int
        Size of the history window.
    prediction_size : int
        Size of the prediction window.
    shift : int
        Shift between consecutive windows.
    history_columns : list
        List of columns for the history window.
    meta_columns : list
        List of columns for the meta data.
    prediction_columns : list
        List of columns for the prediction.
    drop_remainder : bool
        Whether to drop remainder batches.
    history_reference_column : str, optional
        Name of the column containing history reference values.
    meta_reference_column : str, optional
        Name of the column containing meta reference values.
    prediction_reference_column : str, optional
        Name of the column containing prediction reference values.

    Returns
    -------
    int
        Number of batches validated.

    """
    df = df.sort_index()

    if "id" in df.columns:
        ids = len(df.id.unique())
    else:
        ids = 1

    window_size = history_size + prediction_size
    initial_size = window_size - shift
    patch_data = df.index.unique().size - initial_size
    patches = patch_data / shift * ids
    has_remainder = (patches % batch_size != 0) and not drop_remainder
    expected_batches = int(patches // batch_size) + int(has_remainder)

    history_columns = list(sorted(history_columns))
    meta_columns = list(sorted(meta_columns))
    prediction_columns = list(sorted(prediction_columns))

    history_columns_idx = {c: i for i, c in enumerate(history_columns)}
    meta_columns_idx = {c: i for i, c in enumerate(meta_columns)}
    prediction_columns_idx = {c: i for i, c in enumerate(prediction_columns)}

    assert (
        len(set(history_columns + meta_columns + prediction_columns) - set(df.columns))
        == 0
    ), "Not all columns in test df"
    assert (
        history_reference_column in df.columns
    ), f"history_reference_column ({history_reference_column}) not in df.columns"
    assert (
        meta_reference_column in df.columns
    ), f"meta_reference_column ({meta_reference_column}) not in df.columns"
    assert (
        prediction_reference_column in df.columns
    ), f"prediction_reference_column ({prediction_reference_column}) not in df.columns"

    for batch_no, (x, y) in enumerate(ds.as_numpy_iterator(), start=1):
        if batch_no == expected_batches and has_remainder:
            bs = int(patches % batch_size)
        else:
            bs = batch_size
        x1_shape = (bs, history_size, len(history_columns))
        x2_shape = (bs, 1, len(meta_columns))
        y_shape = (bs, prediction_size, len(prediction_columns))

        x1, x2 = None, None
        if history_size and len(history_columns) and len(meta_columns):
            x1, x2 = x
        elif history_size and len(history_columns):
            x1 = x
        elif len(meta_columns):
            x2 = x

        if x1 is not None:
            assert (
                x1.shape == x1_shape
            ), f"Wrong shape: history ({batch_no}: {x1.shape} != {x1_shape})"
            ref = x1[:, 0, history_columns_idx[history_reference_column]]
            assert np.all(
                x1
                == gen_batch(
                    df, history_columns, history_size, ref, history_reference_column
                )
            ), f"Wrong data: history ({batch_no})"
            if x2 is not None:
                assert np.all(
                    x2
                    == gen_batch(
                        df,
                        meta_columns,
                        1,
                        ref + history_size,
                        history_reference_column,
                    )
                ), f"wrong data: meta not consecutive ({batch_no})"

            last_val = x1[:, -1, history_columns_idx[history_reference_column]]
            y_test = gen_batch(
                df,
                prediction_columns,
                prediction_size,
                last_val + 1,
                history_reference_column,
            )
            assert np.all(
                y == y_test
            ), f"Wrong data: prediction not consecutive ({batch_no})"

        if x2 is not None:
            first_val = x2[:, 0, meta_columns_idx[meta_reference_column]]
            assert x2.shape == x2_shape, f"Wrong shape: meta ({batch_no})"
            assert np.all(
                x2 == gen_batch(df, meta_columns, 1, first_val, meta_reference_column)
            ), f"Wrong data: meta ({batch_no})"

        assert y.shape == y_shape, f"Wrong shape: prediction ({batch_no})"
        first_val = y[:, 0, prediction_columns_idx[prediction_reference_column]]
        assert np.all(
            y
            == gen_batch(
                df,
                prediction_columns,
                prediction_size,
                first_val,
                prediction_reference_column,
            )
        ), f"Wrong data: prediction ({batch_no})"

    assert batch_no is not None, "No iteration. Is there enough data?"

    assert (
        batch_no == expected_batches
    ), f"Not enough batches. {batch_no} != {expected_batches}"

    return batch_no


def validate_args(
    prediction_size: int,
    history_columns: List[str],
    meta_columns: List[str],
    history_size: int,
    prediction_columns: List[str],
) -> ContextManager[Union[None]]:
    """Validate input arguments of the PatchPreprocessor class.

    Parameters
    ----------
    prediction_size : int
        The size of the prediction.
    history_columns : List[str]
        List of history columns.
    meta_columns : List[str]
        List of meta columns.
    history_size : int
        The size of the history.
    prediction_columns : List[str]
        List of prediction columns.

    Returns
    -------
    ctxmgr : contextlib.ContextManager
        Context manager indicating if input is valid or raises an AssertionError.

    """
    if prediction_size <= 0:
        ctxmgr = pytest.raises(
            AssertionError,
            match="prediction_size must be a positive integer greater than zero",
        )
    elif len(set(history_columns + meta_columns)) == 0:
        ctxmgr = pytest.raises(
            AssertionError,
            match="No feature columns provided",
        )
    elif len(meta_columns) == 0 and history_size <= 0:
        ctxmgr = pytest.raises(
            AssertionError,
            match="history_size must be a positive integer greater than zero,"
            " when no meta date is used",
        )
    elif history_size < 0:
        ctxmgr = pytest.raises(
            AssertionError,
            match="history_size must be a positive integer",
        )
    elif len(prediction_columns) == 0:
        ctxmgr = pytest.raises(
            AssertionError,
            match="No prediction columns provided",
        )
    else:
        ctxmgr = nullcontext()
    return ctxmgr



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : __init__.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-15 17:18:03 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
from .cyclical_feature_encoder import CyclicalFeatureEncoder  # noqa: F401
from .groupby_dataset_generator import GroupbyDatasetGenerator  # noqa: F401
from .time_series_k_fold import TimeSeriesKFold  # noqa: F401
from .time_series_split import TimeSeriesSplit  # noqa: F401
